<?php

declare(strict_types=1);

namespace Drupal\responsive_bg_image_formatter\Plugin\Field\FieldFormatter;

use Drupal\bg_image_formatter\Plugin\Field\FieldFormatter\BgImageFormatter;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\Theme\Registry;
use Drupal\Core\Url;
use Drupal\responsive_image\Entity\ResponsiveImageStyle;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Responsive Background Image formatter.
 *
 * @FieldFormatter(
 *     id="responsive_bg_image_formatter",
 *     label=@Translation("Responsive Background Image"),
 *     field_types={"image"}
 * )
 */
class ResponsiveBgImageFormatter extends BgImageFormatter {

  /**
   * The theme registry service.
   *
   * @var \Drupal\Core\Theme\Registry
   */
  protected Registry $registry;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->registry = $container->get('theme.registry');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $element = parent::settingsForm($form, $form_state);
    $element['image_style']['#options'] = $this->getResponsiveImageStyles(TRUE);
    $element['image_style']['#description'] = $this->t(
      'Select <a href="@href_image_style">the responsive image style</a> to use.',
      [
        '@href_image_style' => Url::fromRoute('entity.responsive_image_style.collection')->toString(),
      ]
    );

    unset($element['css_settings']['bg_image_media_query']);

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = parent::settingsSummary();
    $settings = $this->getSettings();
    $options = $this->getResponsiveImageStyles();

    if (isset($options[$settings['image_style']])) {
      $summary[1] = $this->t('URL for image style: @style', ['@style' => $options[$settings['image_style']]]);
    }
    else {
      $summary[1] = $this->t('Original image style');
    }

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $elements = [];
    $settings = $this->getSettings();
    $css_settings = $settings['css_settings'];
    $selectors = array_filter(preg_split('/$/', $css_settings['bg_image_selector']));
    $files = $this->getEntitiesToView($items, $langcode);

    // Early opt-out if the field is empty.
    if (empty($files) || empty($settings['image_style'])) {
      return $elements;
    }

    // Create simpler field identifier.
    $field_id = $items->getName();

    // Initialize selector index if not set.
    static::$selectorIndex[$field_id] ??= 0;

    // Prepare token data in bg image css selector.
    $token_data = [
      'user' => $this->currentUser,
      $items->getEntity()->getEntityTypeId() => $items->getEntity(),
    ];

    // Clean and resolve selectors.
    $selectors = array_map(function ($selector) use ($token_data) {
      $selector = trim($selector, ', ');
      return $this->token->replace($selector, $token_data);
    }, $selectors);

    // Need an empty element so views renderer will see something to render.
    $elements[0] = [];

    /** @var \Drupal\file\FileInterface $file */
    foreach ($files as $delta => $file) {
      // Use the static index for selector cycling.
      $selector = !empty($selectors) ? $selectors[static::$selectorIndex[$field_id] % \count($selectors)] : 'body';
      // Increment the index after using it.
      static::$selectorIndex[$field_id]++;

      $uri = $file->getFileUri();
      $variables = [
        'uri' => $file->getFileUri(),
        'responsive_image_style_id' => $settings['image_style'],
        'width' => 1,
        'height' => 1,
      ];
      if (file_exists($uri)) {
        [$variables['width'], $variables['height']] = getimagesize($uri);
      }

      // Call preprocess functions to let other modules/themes react before
      // rendering.
      $info = $this->registry->get();
      $hook = 'responsive_image__responsive_bg_image_formatter';
      foreach ($info['responsive_image']['preprocess functions'] as $preprocessor_function) {
        if (is_callable($preprocessor_function)) {
          call_user_func_array($preprocessor_function, [&$variables, $hook, $info]);
        }
      }

      if (empty($variables['sources'])) {
        continue;
      }

      // Split each source into multiple rules.
      foreach (array_reverse($variables['sources']) as $source_i => $source) {
        $attr = $source->toArray();

        $srcset = explode(', ', $attr['srcset']);

        foreach ($srcset as $src_i => $src) {
          [$src, $res] = explode(' ', $src);

          $media = $attr['media'] ?? '';

          // Add "retina" to media query if this is a 2x image.
          if ($res && $res === '2x') {
            $media = "{$media} and (-webkit-min-device-pixel-ratio: 2), {$media} and (min-resolution: 192dpi)";
          }

          // Correct a bug in template_preprocess_responsive_image which
          // generates an invalid media rule "screen (max-width)" when no
          // min-width is specified. If this bug gets fixed, this replacement
          // will deactivate.
          $media = str_replace('screen (max-width', 'screen and (max-width', $media);

          $css_settings['bg_image_selector'] = $selector;
          $css = $this->getBackgroundImageCss($src, $css_settings);
          $css['media'] = $media;

          // Define unique key to prevent collisions when displaying multiple
          // background images on the same page.
          $html_head_key = 'responsive_bg_image_formatter_css__' . sha1(
            implode('__', [
              $items->getEntity()->uuid(),
              $items->getName(),
              $settings['image_style'],
              $delta,
              $src_i,
              $source_i,
            ])
          );

          $style_element = [
            '#type' => 'html_tag',
            '#tag' => 'style',
            '#attributes' => [
              'media' => $css['media'],
            ],
            '#value' => Markup::create($css['style']),
          ];

          if ($this->isAjax() || $this->request->isXmlHttpRequest()) {
            $elements['#attached']['drupalSettings']['bg_image_formatter_css'][$html_head_key] =
            $this->renderer->renderInIsolation($style_element);
          }
          else {
            $elements['#attached']['html_head'][] = [$style_element, $html_head_key];
          }
        }
      }
    }

    return $elements;
  }

  /**
   * Get the possible responsive image styles.
   *
   * @param bool $withNone
   *   True to include the 'None' option, false otherwise.
   *
   * @return array
   *   The select options.
   */
  protected function getResponsiveImageStyles($withNone = FALSE) {
    $styles = ResponsiveImageStyle::loadMultiple();
    $options = [];

    if ($withNone && empty($styles)) {
      $options[''] = $this->t('- Defined None -');
    }

    foreach ($styles as $name => $style) {
      $options[$name] = $style->label();
    }

    return $options;
  }

}
